@@ -10,15 +10,16 @@ Without browser_kill(), your browser will stay opened until you close it in your
 © Anime no Sekai - 2020
 """
 
-import os
 import warnings
 import psutil
-from lifeeasy import write_file, today, current_time, find_inside_file
+from lifeeasy import write_file, today, current_time
 
 from selenium import webdriver
 from selenium.common.exceptions import NoSuchElementException
 
-import google_translate_data
+from internal.caching import search_translation_cache, add_translation_cache
+from internal.language_code import verify_language_code
+
 
 class BrowserError(Exception):
     """
@@ -30,99 +31,13 @@ class BrowserError(Exception):
         exception_msg = f"\n\n⚠️ ⚠️ ⚠️\n{self.msg}\n"
         return exception_msg
 
-class GoogleTranslateDomainError(Exception):
-    """
-    When the domain is incorrect.
-    """
-    def __init__(self, msg=None):
-        self.msg = msg 
-    def __str__(self):
-        exception_msg = f"\n\n⚠️ ⚠️ ⚠️\n{self.msg}\n"
-        return exception_msg
-
-class LanguageCodeError(Exception):
-    """
-    When the language code is incorrect.
-    """
-    def __init__(self, msg=None):
-        self.msg = msg 
-    def __str__(self):
-        exception_msg = f"\n\n⚠️ ⚠️ ⚠️\n{self.msg}\n"
-        return exception_msg
 
 warnings.filterwarnings('ignore')
 
 driver_name = ''
 driver = None
-gt_domain = 'translate.google.com'
 connected = False
 last_translation = ''
-translation_caches_path = os.path.dirname(os.path.abspath(__file__)) + '/translation_caches/'
-translation_caches_name = 'translation_caches.animenosekai_caches'
-
-def google_translate_domain(domain='translate.google.com'):
-    """
-    Changes the used google translate domain name (i.e translate.google.com) to the given domain (i.e translate.google.co.jp)
-    """
-    global gt_domain
-    clean_domain = domain.lower().replace('https://', '').replace('http://', '').replace('www.', '').split('/')[0]
-    if clean_domain in google_translate_data.google_translate_domains_data():
-        gt_domain = clean_domain
-        return gt_domain
-    else:
-        raise GoogleTranslateDomainError(f'{clean_domain} is not a correct Google Translate domain.\n {gt_domain} will be used instead.')
-
-def verify_language_code(language):
-    """
-    Verifies and gets the correct language code.
-    """
-    if language.lower() == 'zh-cn':
-        return 'zh-CN'
-    if language.lower() == 'zh-tw':
-        return 'zh-TW'
-    if language.lower() in google_translate_data.language_code_data():
-        return language.lower()
-    else:
-        try:
-            result = google_translate_data.language_name_to_code()[language.lower()]
-            return result
-        except:
-            try:
-                result = google_translate_data.international_language_name_to_code()[language.lower()]
-                return result
-            except:
-                raise LanguageCodeError(f'{language} is not a correct language code.')
-
-def search_translation_cache(source_language, destination_language, source):
-    """
-    Searches a translation through the caches.
-    """
-    line_start = f'[sl={source_language}]｜[dl={destination_language}]｜[source={source}]'
-    search_result = find_inside_file(translation_caches_path + translation_caches_name, line_start)
-    if not search_result == {}:
-        return {'result': search_result['line'].split('｜')[3][8:][:-2], 'line_number': str(search_result['line_number'])}
-    else:
-        line_start = f'[sl={destination_language}]｜[dl={source_language}]'
-        search_result = find_inside_file(translation_caches_path + translation_caches_name, line_start, whole_document=True)
-        if not search_result == {}:
-            found = False
-            for result in search_result:
-                if result['line'].split('｜')[3][8:][:-2] == source:
-                    found = True
-                    return {'result': result['line'].split('｜')[2][8:][:-1], 'line_number': str(result['line_number'])}
-            if not found:
-                return None
-        else:
-            return None
-
-def add_translation_cache(source_language, destination_language, source, result):
-    """
-    Adds a translation to the cache file.
-    """
-    if source_language == 'auto':
-        detect_language(source)
-    line = f'[sl={source_language}]｜[dl={destination_language}]｜[source={source}]｜[result={result}]\n'
-    write_file(translation_caches_name, line, translation_caches_path, append=True)
 
 def browser(browser_name, executable_path="PATH"):
     """
@@ -139,7 +54,7 @@ def browser(browser_name, executable_path="PATH"):
     global driver
     global driver_name
     global connected
-    if not driver is None:
+    if connected:
         browser_kill()
     if browser_name.lower() == 'firefox':
         from selenium.webdriver.firefox.options import Options
@@ -209,6 +124,7 @@ def translate(text, destination_language, source_language="auto", cache=False, d
     Returns a string with the text translated.\n
     Returns "An error occured while translating: translation not found." if the translation was not found in the webpage. This might come from a mistyped language code.
     """
+    from internal.domain import gt_domain
     global last_translation
     if debug:
         write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Starting Translation...\n', append=True)
@@ -220,110 +136,143 @@ def translate(text, destination_language, source_language="auto", cache=False, d
             line_number = cache_result['line_number']
             write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Translation found in Caches (line {line_number})\n', append=True)
         return cache_result['result']
-    if driver is None:
-        if debug:
-            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - No driver selected\n', append=True)
-        raise BrowserError("Browser is not set yet.\n Please set it with browser()")
-    if not connected:
-        if debug:
-            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Driver disconnected, last driver: {driver_name}\n', append=True)
-        raise BrowserError(f'You disconnected the last browser in use ({driver_name}).\n Please reconnect one with browser()')
     else:
-        if debug:
-            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - First attempt url is: https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}\n', append=True)
-        driver.get(f"https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}")
-        try:
+        if driver is None:
             if debug:
-                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Getting DOM Element by Class Name (tlid-translation)\n', append=True)
-            result = driver.find_element_by_class_name("tlid-translation")
-            if result.text == last_translation or result.text == str(last_translation + '...'):
-                if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Translation not finished detected... Refreshing page before new attempt...\n', append=True)
-                driver.refresh()
+                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - No driver selected\n', append=True)
+            raise BrowserError("Browser is not set yet.\n Please set it with browser()")
+        if not connected:
+            if debug:
+                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Driver disconnected, last driver: {driver_name}\n', append=True)
+            raise BrowserError(f'You disconnected the last browser in use ({driver_name}).\n Please reconnect one with browser()')
+        else:
+            if debug:
+                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - First attempt url is: https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}\n', append=True)
+            driver.get(f"https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}")
+            try:
                 if debug:
                     write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Getting DOM Element by Class Name (tlid-translation)\n', append=True)
                 result = driver.find_element_by_class_name("tlid-translation")
-            if debug:
-                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Setting last_translation global variable to new translation...\n', append=True)
-            last_translation = str(result.text)
-            if cache:
-                if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Adding result to cache...\n', append=True)
-                add_translation_cache(source_language=source_language, destination_language=destination_language, source=text, result=str(result.text))
-            if debug:
-                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Returning value... {result.text}\n', append=True)
-            return str(result.text)
-        except NoSuchElementException:
-            if debug:
-                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Element not found on page...\n', append=True)
-            try:
-                if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] New attempt...\n', append=True)
-                driver.get(f"https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}")
-                if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
-                result = driver.find_element_by_class_name("tlid-translation")
                 if result.text == last_translation or result.text == str(last_translation + '...'):
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Translation not finished detected... Refreshing page before new attempt...\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Translation not finished detected... Refreshing page before new attempt...\n', append=True)
                     driver.refresh()
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Getting DOM Element by Class Name (tlid-translation)\n', append=True)
                     result = driver.find_element_by_class_name("tlid-translation")
                 if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Setting last_translation global variable to new translation...\n', append=True)
+                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Setting last_translation global variable to new translation...\n', append=True)
                 last_translation = str(result.text)
                 if cache:
                     if debug:
                         write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Adding result to cache...\n', append=True)
                     add_translation_cache(source_language=source_language, destination_language=destination_language, source=text, result=str(result.text))
                 if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Returning value... {result.text}\n', append=True)
+                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Returning value... {result.text}\n', append=True)
                 return str(result.text)
             except NoSuchElementException:
                 if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Element not found on page...\n', append=True)
+                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Element not found on page...\n', append=True)
                 try:
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] New attempt...\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] New attempt...\n', append=True)
                     driver.get(f"https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}")
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Translation not finished detected... Refreshing page before new attempt...\n', append=True)
-                    driver.refresh()
-                    if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
                     result = driver.find_element_by_class_name("tlid-translation")
+                    if result.text == last_translation or result.text == str(last_translation + '...'):
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Translation not finished detected... Refreshing page before new attempt...\n', append=True)
+                        driver.refresh()
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
+                        result = driver.find_element_by_class_name("tlid-translation")
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Setting last_translation global variable to new translation...\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Setting last_translation global variable to new translation...\n', append=True)
                     last_translation = str(result.text)
                     if cache:
                         if debug:
                             write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Adding result to cache...\n', append=True)
                         add_translation_cache(source_language=source_language, destination_language=destination_language, source=text, result=str(result.text))
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Returning value... {result.text}\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Returning value... {result.text}\n', append=True)
                     return str(result.text)
                 except NoSuchElementException:
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Element not found, aborting...\n', append=True)
-                    return "An error occured while translating: translation not found."
-                except:
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Element not found on page...\n', append=True)
+                    try:
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] New attempt...\n', append=True)
+                        driver.get(f"https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}")
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Translation not finished detected... Refreshing page before new attempt...\n', append=True)
+                        driver.refresh()
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
+                        result = driver.find_element_by_class_name("tlid-translation")
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Setting last_translation global variable to new translation...\n', append=True)
+                        last_translation = str(result.text)
+                        if cache:
+                            if debug:
+                                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Adding result to cache...\n', append=True)
+                            add_translation_cache(source_language=source_language, destination_language=destination_language, source=text, result=str(result.text))
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Returning value... {result.text}\n', append=True)
+                        return str(result.text)
+                    except NoSuchElementException:
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Element not found, aborting...\n', append=True)
+                        return "An error occured while translating: translation not found."
+                    except Exception as e:
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Unknown error\n', append=True)
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Error details: {str(e)}\n', append=True)
+                        return "An error occured while translating: unknown error."
+                except Exception as e:
                     if debug:
-                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 3] Unknown error\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Unknown error\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Error details: {str(e)}\n', append=True)
                     return "An error occured while translating: unknown error."
-            except:
+            except Exception as e:
                 if debug:
-                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Unknown error\n', append=True)
-                return "An error occured while translating: unknown error."
-        except:
-            if debug:
-                write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Unknown error\n', append=True)
-            return "An error occured while translating: unknown error."
+                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Unknown error\n', append=True)
+                    write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Error details: {str(e)}\n', append=True)
+                try:
+                    if debug:
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] New attempt...\n', append=True)
+                    driver.get(f"https://{gt_domain}/?hl=en#view=home&op=translate&sl={verify_language_code(source_language)}&tl={verify_language_code(destination_language)}&text={str(text)}")
+                    if debug:
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
+                    result = driver.find_element_by_class_name("tlid-translation")
+                    if result.text == last_translation or result.text == str(last_translation + '...'):
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Translation not finished detected... Refreshing page before new attempt...\n', append=True)
+                        driver.refresh()
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Getting DOM Element by Class Name (tlid-translation)\n', append=True)
+                        result = driver.find_element_by_class_name("tlid-translation")
+                    if debug:
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Setting last_translation global variable to new translation...\n', append=True)
+                    last_translation = str(result.text)
+                    if cache:
+                        if debug:
+                            write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Adding result to cache...\n', append=True)
+                        add_translation_cache(source_language=source_language, destination_language=destination_language, source=text, result=str(result.text))
+                    if debug:
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - [Attempt 2] Returning value... {result.text}\n', append=True)
+                    return str(result.text)
+                except Exception as e:
+                    if debug:
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Unknown error\n', append=True)
+                        write_file('logs.txt', today() + ' ' + current_time() + f'   text={text}｜sl={source_language}｜dl={destination_language} - Error details: {str(e)}\n', append=True)
+                    return "An error occured while translating: unknown error."
 
 def detect_language(text, result_language='en'):
     """
     Returns the language of the given text.
     """
+    from internal.domain import gt_domain
     if driver is None:
         raise BrowserError("Browser is not set yet.\n Please set it with browser()")
     if not connected:
@@ -345,6 +294,7 @@ def transliterate(text, source_language="auto"):
     Returns the transliteration provided by Google Translate (if available)\n
     i.e Ohayou --> おはよう / おはよう --> Ohayou
     """
+    from internal.domain import gt_domain
     if driver is None:
         raise BrowserError("Browser is not set yet.\n Please set it with browser()")
     if not connected:
@@ -362,6 +312,7 @@ def definition(text, source_language="auto"):
     """
     Returns the word type (i.e Interjection, Noun), defintion (if available) and sentence example where the word could be used (if available)
     """
+    from internal.domain import gt_domain
     if driver is None:
         raise BrowserError("Browser is not set yet.\n Please set it with browser()")
     if not connected:
